import random
import re
import threading

from jardin import config as config
from jardin.database.database_config import DatabaseConfig


class Datasources(object):

    class IsolatedDbClients(threading.local):
        def __init__(self) -> None:
            # All clients indexed by db_name.
            self.all = {}  # type: dict[str, list[BaseClient]]

            # Active clients indexed by db_name. An active client is a sticky connection for the
            # life of a session. The application can reset a session by calling `self.shuffle_clients()`
            # at an appropriate moment in the app lifecycle (e.g. the beginning of a job or request).
            self.active = {}  # type: dict[str, BaseClient]

    # Each thread will have its own cache of active database clients/connections
    _clients = IsolatedDbClients()

    # Mapping from db_name to a list of config info. The list will typically contain one element,
    # but in the case of a DB with many replicas, there will be one config object for each replica.
    # Read-only and shared across threads.
    _db_configs = {}  # type: dict[str, list[DatabaseConfig]]

    # Guards lazy initializers
    _lock = threading.Lock()

    SUPPORTED_SCHEMES = ('postgres', 'mysql', 'sqlite', 'snowflake', 'redshift')

    @classmethod
    def active_client(self, db_name):
        if db_name not in self._clients.active:
            clients = self._clients.all.get(db_name)
            if clients is None:
                clients = self._build_clients(db_name)
                self._clients.all[db_name] = clients
            c = clients[0] if len(clients) == 1 else random.choice(clients)
            self.log_datasource(db_name, c.db_config)
            self._clients.active[db_name] = c
        return self._clients.active[db_name]

    @classmethod
    def _build_clients(self, name):
        clients = []
        configs = self.db_configs(name)
        for db in configs:
            if db.scheme not in self.SUPPORTED_SCHEMES:
                raise UnsupportedDatabase('%s is not a supported database' % db.scheme)
            elif db.scheme == 'postgres' or db.scheme == 'redshift':
                import jardin.database.clients.pg as impl
            elif db.scheme == 'mysql':
                import jardin.database.clients.mysql as impl
            elif db.scheme == 'sqlite':
                import jardin.database.clients.sqlite as impl
            elif db.scheme == 'snowflake':
                import jardin.database.clients.sf as impl
            clients.append(impl.DatabaseClient(db, name))
        return clients

    @classmethod
    def db_configs(self, name):
        if not self._db_configs:
            with self._lock:
                if not self._db_configs:
                    self._db_configs = self._build_db_configs()
        return self._db_configs[name]

    @classmethod
    def _build_db_configs(self):
        config.init()
        d = dict()
        for (db_name, val) in config.DATABASES.items():
            # we don't support multi-configs of dictionary format yet; the "else [urls]" is for a dictionary
            url_list = re.split(r'\s+', val) if isinstance(val, str) else [val]
            d[db_name] = [DatabaseConfig(x) for x in url_list]
        return d

    @classmethod
    def shuffle_clients(self):
        for name, clients in self._clients.all.items():
            c = None
            if len(clients) == 1:
                c = clients[0]
            else:
                active = self._clients.active[name]
                filtered = list(filter(lambda x: x is not active, clients))
                c = filtered[0] if len(filtered) == 1 else random.choice(filtered)
            self.log_datasource(name, c.db_config)
            self._clients.active[name] = c

    @classmethod
    def log_datasource(self, name, db_config):
        host = getattr(db_config, 'host', None) or '_'  # use "_" for both missing attr or None value cases
        port = getattr(db_config, 'port', None) or '_'
        user = getattr(db_config, 'username', None) or '_'
        database = getattr(db_config, 'database', None) or '_'
        config.logger.debug("[{}]: datasource {}@{}:{}/{}".format(name, user, host, port, database))


class UnsupportedDatabase(Exception): pass


import pandas

from jardin import config as config
from jardin.query_builders import \
    SelectQueryBuilder, \
    InsertQueryBuilder, \
    UpdateQueryBuilder, \
    DeleteQueryBuilder, \
    RawQueryBuilder
from jardin.cache_stores import cached


def set_defaults(func):
    def wrapper(self, *args, **kwargs):
        kwargs.update(
            model_metadata=self.model_metadata,
            scheme=self.db_client.db_config.scheme,
            lexicon=self.db_client.lexicon
            )
        return func(self, *args, **kwargs)
    return wrapper


class DatabaseAdapter(object):

    def __init__(self, db_client, model_metadata):
        self.db_client = db_client
        self.model_metadata = model_metadata

    @set_defaults
    def select(self, **kwargs):
        query = SelectQueryBuilder(**kwargs).query
        config.logger.debug(query)
        results, columns = self.db_client.execute(*query, write=False)
        if results is None and columns is None:
            return None
        return pandas.DataFrame.from_records(results, columns=columns, coerce_float=True)

    @set_defaults
    def write(self, query_builder, **kwargs):
        query = query_builder(**kwargs).query
        config.logger.debug(query)
        returning_ids = self.db_client.execute(*query, write=True, **kwargs)
        if len(returning_ids) > 0:
            return self.select(where={kwargs['primary_key']: returning_ids})
        return None

    def insert(self, **kwargs):
        return self.write(InsertQueryBuilder, **kwargs)

    def update(self, **kwargs):
        return self.write(UpdateQueryBuilder, **kwargs)

    @set_defaults
    def delete(self, **kwargs):
        query = DeleteQueryBuilder(**kwargs).query
        config.logger.debug(query)
        self.db_client.execute(*query, write=False)

    @set_defaults
    @cached
    def raw_query(self, **kwargs):
        query = RawQueryBuilder(**kwargs).query
        config.logger.debug(query)
        results, columns = self.db_client.execute(*query, write=False)
        if results is None and columns is None:
            return None
        return pandas.DataFrame.from_records(results, columns=columns, coerce_float=True)



import sqlite3

from jardin.database.base_client import BaseClient
from jardin.database.base_lexicon import BaseLexicon


class Lexicon(BaseLexicon):

    @staticmethod
    def table_schema_query(table_name):
        return "pragma table_info(%s);" % table_name

    @staticmethod
    def column_info(row):
        return row['name'], row['dflt_value'], row['type']

    @staticmethod
    def extrapolator(field):
        return ':%s' % field

    @staticmethod
    def row_ids(cursor, primary_key):
        return [cursor.lastrowid]


class DatabaseClient(BaseClient):

    lexicon = Lexicon
    retryable_exceptions = tuple()

    def connect_impl(self):
        # autocommit is enabled by setting isolation_level to None
        return sqlite3.connect(self.db_config.database, isolation_level=None)

    def execute_impl(self, conn, *query):
        cursor = conn.cursor()
        cursor.execute(*query)
        return cursor



import time
from abc import ABC, abstractmethod

MAX_RETRIES = 3


class BaseClient(ABC):

    def __init__(self, db_config, name):
        self.db_config = db_config
        self.name = name
        self._conn = None
        
    @property
    def default_connect_kwargs(self):
        return dict(
            database=self.db_config.database,
            user=self.db_config.username,
            password=self.db_config.password,
            host=self.db_config.host,
            port=self.db_config.port,
            connect_timeout=5
        )

    @property
    @abstractmethod
    def lexicon(self):
        """Provide an object which normalizes a SQL dialect."""\

    @property
    @abstractmethod
    def retryable_exceptions(self):
        """Provide exceptions which may be retried."""

    @abstractmethod
    def connect_impl(self):
        """Connect to a SQL database."""

    @abstractmethod
    def execute_impl(self, conn, *query):
        """Execute a SQL query and return the cursor."""

    def execute(self, *query, **kwargs):
        last_exception = None
        delay = 3
        for _ in range(MAX_RETRIES):
            try:
                return self.execute_once(*query, **kwargs)
            except self.retryable_exceptions as e:
                # TODO [kl] comment this out?
                print(f"{e}, Retrying in {delay} seconds...")
                time.sleep(delay)
                delay *= 2
                last_exception = e
                continue
        else:
            raise last_exception

    def execute_once(self, *query, write=False, **kwargs):
        """Connect to the database (if necessary) and execute a query."""
        if self._conn is None:
            self._conn = self.connect_impl()

        try:
            cursor = self.execute_impl(self._conn, *query)
        except self._conn.InterfaceError:
            # the connection is probably closed; stop re-using the connection
            self._conn = None
            raise

        if write:
            return self.lexicon.row_ids(cursor, kwargs['primary_key'])
        if cursor.description:
            return cursor.fetchall(), self.columns(cursor)
        return None, None

    def columns(self, cursor):
        cursor_desc = cursor.description
        columns = []
        if cursor_desc:
            columns = [col_desc[0] for col_desc in cursor_desc]
            if self.db_config.lowercase_columns:
                columns = [col.lower() for col in columns]
        return columns


import imp, os, logging, sys

DEFAULTS = {
    'WATERMARK': '',
    'LOG_LEVEL': logging.INFO,
    'CACHE': {
        'method': None,
        'methods': {
            's3': {
                'bucket_name': None,
                'path': ''
            },
            'disk': {
                'dir': None,
                'limit': None
            }
        }
    }
}

INITIALIZED = False

def init():
    if INITIALIZED:
        return
    
    global DATABASES, CACHE, WATERMARK, LOG_LEVEL, logger
    
    config_file = imp.load_source('jardin_conf', os.environ.get('JARDIN_CONF', 'jardin_conf.py'))
    
    DATABASES = config_file.DATABASES
    
    this = sys.modules[__name__]
    
    for (k, v) in DEFAULTS.items():
        if hasattr(config_file, k):
            v = getattr(config_file, k)
        setattr(this, k, v)
    
    logging.basicConfig(level=LOG_LEVEL)
    logger = logging.getLogger('jardin')
    logger.setLevel(LOG_LEVEL)
    
    setattr(this, 'INITIALIZED', True)



import argparse
import os
import inspect

from jardin.database.database_adapter import DatabaseAdapter
from jardin.database.datasources import Datasources
from jardin.tools import stack_marker

def query(sql=None, filename=None, extract=None, db=None, **kwargs):
    if db is None:
        raise argparse.ArgumentError('You must provide a database name')

    kwargs['stack'] = stack_marker(inspect.stack())

    filename = filename or extract

    if filename and not filename.startswith('/'):
        filename = os.path.join(os.environ['PWD'], filename)

    if 'where' not in kwargs and 'params' in kwargs:
        kwargs['where'] = kwargs['params']

    return DatabaseAdapter(
        Datasources.active_client(db),
        None
        ).raw_query(
            sql=sql, filename=filename, **kwargs
        )



import psycopg2 as pg
from psycopg2 import extras

from jardin.database.base_client import BaseClient
from jardin.database.base_lexicon import BaseLexicon
import jardin.config as config


class Lexicon(BaseLexicon):

    @staticmethod
    def table_schema_query(table_name):
        return "SELECT column_name, column_default, data_type FROM " \
            "information_schema.columns WHERE " \
            "table_name=%(table_name)s AND table_schema='public';"

    @staticmethod
    def column_info(row):
        return row['column_name'], row['column_default'], row['data_type']

    @staticmethod
    def update_values(fields, value_extrapolators):
        if len(fields) == 1:
            result = ', '.join(fields) + ' = ' + \
                [', '.join(ext)
                 for ext in value_extrapolators][0]
        else:
            result = '(' \
                + ', '.join(fields) \
                + ') = '
            result += ', '.join(
                ['(' + ', '.join(ext) + ')' for ext in value_extrapolators]
            )
        return result

    @staticmethod
    def row_ids(cursor, primary_key):
        row_ids = cursor.fetchall()
        return [r[primary_key] for r in row_ids]


class DatabaseClient(BaseClient):

    lexicon = Lexicon
    retryable_exceptions = (pg.OperationalError, pg.InterfaceError, pg.extensions.QueryCanceledError)

    def connect_impl(self):
        kwargs = self.default_connect_kwargs.copy()
        kwargs.update(connection_factory=extras.MinTimeLoggingConnection)
        conn = pg.connect(**kwargs)
        conn.initialize(config.logger)
        conn.autocommit = True
        return conn

    def execute_impl(self, conn, *query):
        cursor = conn.cursor(cursor_factory=pg.extras.RealDictCursor)
        cursor.execute(*query)
        return cursor



import pymysql

from jardin.database.base_client import BaseClient
from jardin.database.base_lexicon import BaseLexicon


class Lexicon(BaseLexicon):

    @staticmethod
    def table_schema_query(table_name):
        return "SHOW COLUMNS FROM %s;" % table_name

    @staticmethod
    def column_info(row):
        return row['Field'], row['Default'], row['Type']

    @staticmethod
    def row_ids(cursor, primary_key):
        cursor.execute('SELECT LAST_INSERT_ID();')
        return [cursor.fetchall()[0][0]]

    @staticmethod
    def apply_watermark(query, watermark):
        return ' '.join([watermark, query])


class DatabaseClient(BaseClient):

    lexicon = Lexicon
    retryable_exceptions = (pymysql.InterfaceError, pymysql.OperationalError)

    def connect_impl(self):
        kwargs = self.default_connect_kwargs.copy()
        kwargs.update(autocommit=True)
        return pymysql.connect(**kwargs)

    def execute_impl(self, conn, *query):
        cursor = conn.cursor()
        cursor.execute(*query)
        return cursor



import re

import snowflake.connector as sf

from jardin.database.clients.pg import Lexicon as PGLexicon
from jardin.database.base_client import BaseClient


class Lexicon(PGLexicon):

    @staticmethod
    def extrapolator(_):
        return '%s'

    @staticmethod
    def format_args(args):
        return args.values()

    @staticmethod
    def standardize_interpolators(sql, params):
        sql, params = super(
            Lexicon, Lexicon).standardize_interpolators(sql, params)
        param_names = re.findall(r'\%\((\w+)\)s', sql)
        if len(param_names):
            sql = re.sub(r'\%\(\w+\)s', '%s', sql)
            if isinstance(params, dict):
                params = list(map(lambda x: params[x], param_names))
        return sql, params


class DatabaseClient(BaseClient):

    lexicon = Lexicon
    retryable_exceptions = (sf.InterfaceError, sf.OperationalError)

    def connect_impl(self):
        kwargs = dict(
            user=self.db_config.username,
            password=self.db_config.password,
            account=self.db_config.account,
            database=self.db_config.database,
            schema=self.db_config.schema,
            autocommit=True
        )
        if 'warehouse' in dir(self.db_config):
            kwargs['warehouse'] = self.db_config.warehouse
        if 'authenticator' in dir(self.db_config):
            kwargs['authenticator'] = self.db_config.authenticator
        if 'client_session_keep_alive' in dir(self.db_config):
            kwargs['client_session_keep_alive'] = self.db_config.client_session_keep_alive

        return sf.connect(**kwargs)

    def execute_impl(self, conn, *query):
        cursor = conn.cursor()
        cursor.execute(*query)
        return cursor
